{Feedback Servo Driver}

CON
        _clkmode = xtal1 + pll16x    'Standard clock mode 
        _xinfreq = 5_000_000         '* crystal frequency = 80 MHz
        
VAR                                       
  long  MyParm[10]                'passing addresses, etc to pasm cog
  byte  Cog
  word  PulseWidth[4]                 'control pulse width in microseconds
  word  theta[4]                      'feedback angle in degrees
  byte  mode[4]                       '0 => velocity mode, <>0 => position mode
  word  commanded[4]                  'commanded angle               ' 
  byte  Complete[4]                   '0 => servo has completed move
  word  ServoPins[4]                  'ControlPin<<8 | FeedbackPin
  word  GoodCount[4]                  'used by pasm cog
  word  index
  
PUB start(CP0,FP0,CP1,FP1,CP2,FP2,CP3,FP3 ) 'start the control cog
  ServoPins[0] := CP0 << 8 | FP0         'set up the pin numbers
  ServoPins[1] := CP1 << 8 | FP1
  ServoPins[2] := CP2 << 8 | FP2
  ServoPins[3] := CP3 << 8 | FP3
  repeat index from 0 to 3
    Mode[index] := 0                     'velocity mode       
    PulseWidth[index] := 1500            'start up still
  MyParm[0] := @PulseWidth               'address of pulse width to motor
  MyParm[1] := @Mode                     '0 => velocity mode, <>0 => position mode
  MyParm[2] := @theta                    'feedback: actual angle
  MyParm[3] := @Commanded                'commanded angle
  MyParm[4] := @Complete                 'done flag
  MyParm[5] := clkfreq / 1_000_000       'tics in 1 microseconds               
  MyParm[6] := @ServoPins                'pin numbers
  MyParm[7] := @GoodCount                'used by pasm cog 
  cog := cognew(@BCog, @MyParm)           'start up the pasm cog
  return cog            

PUB SetPos(servo, degrees)
    Commanded[Servo] := degrees           'save commanded angle
    Complete[Servo] := 1                  'we are not complete yet
    Mode[Servo] := 1                      'set position mode
    
PUB Width(servo, Usec)                     'set pulse width
    mode[servo] := 0                       'velocity mode
    pulsewidth[servo] := USec               'save pulse width (in usec)

PUB GetPos(servo)                        'get current position
    return theta[servo]

PUB CheckDone(servo)                     'return zero says position completed
    return Complete[servo]         
DAT   '***********pasm cog******************
BCog    org   0                'drives control pulse, returns current angle               '
        mov AdPar,Par          'get the hub address of control PulseWidth list
        rdlong AdPWidth, AdPar
        add AdPar, #4           'to address of mode list
        rdlong AdMode, AdPar '
        add AdPar, #4           'to address of theta list
        rdlong AdTheta, AdPar
        add AdPar, #4             'to address of commanded angle list
        rdlong AdCommand, AdPar
        add AdPar, #4             'to address of completed flag list
        rdlong AdComplete, AdPar  
        add AdPar, #4             'to constant clk tics in 1 usec
        rdlong MicroSec, AdPar
        add AdPar, #4             'to pin numbers list
        rdlong AdPins, AdPar
        add AdPar, #4             'to goodcount list
        rdlong AdGoodCnt, AdPar  
        mov work1, #%00100          'pina follows phsa[31]
        shl work1, #26               'to control field
        mov ctra, work1             'start up counter a
        mov frqa, #1                'count by one
        or dira, RFBMask            'engineering
        or dira, Corr
        or dira, P3
        or dira, P4
        or dira, P5
        mov x, MicroSec             'tics in one microsecond 
        mov y, C2500                '
        call #multiply              'tics in 2.5 msec
        mov C2500, y                'save for later use
        mov x, MicroSec             
        mov y, C1500                '
        call #multiply              'tics in 1.5 msec
        mov C1500, y                'save for later (check zero crossing)
        mov TheServo, #0            'begin with zero          
        mov timer, cnt              'initialize 15 msec timer
        add timer, C2500
loop    test TheServo, #3 wz        'see if servo number zero
  if_z  jmp #InitAds                'go point to beginnings of lists
        add CAdPWidth, #2           'go on to next set of context (word)
        add CAdMode, #1             'byte
        add CAdTheta, #2            'word
        add CADCommand, #2          'word
        add CAdComplete, #1         'byte
        add CAdPins, #2             'word
        add CAdGoodCnt, #2          'word 
        jmp #GetCon                 'get context
InitAds mov CAdPWidth, AdPWidth     'get pointers for servo zero
        mov CAdMode, AdMode
        mov CAdTheta, AdTheta
        mov CAdCommand, AdCommand
        mov CAdComplete, AdComplete
        mov CAdPins, AdPins
        mov CAdGoodCnt, AdGoodCnt
GetCon  rdword CountGood, CAdGoodCnt  'getting context
        nop                        'may need to do something else here              
        waitcnt timer, C2500       'wait 2.5 msec
        or outa, RFBMask            'scope sync        
        rdword work1, CAdPins wz   'get the pin numbers
   if_Z jmp #NoServo               'no servo here     
        mov ContPin, Work1          'keep the control pin for later
        shr ContPin, #8             'scaled appropriately
        and Work1, #31              'feedback pin
        mov FBPinMask, #1           'form the mask for the feedback pin
        shl FBPinMask, Work1        
        waitpne FBPinMask, FBPinMask 'wait for feedback low
        waitpeq FBPinMask, FBPinMask 'then high
        mov work1, cnt               'beginning of high feedback pulse
        waitpne FBPinMask, FBPinMask 'wait for end of high pulse
        mov HiPer, cnt
        waitpeq FBPinMask, FBPinMask 'now wait for end of low pulse
        mov LowPer, cnt
        sub LowPer, HiPer           'calculate low period
        sub HiPer, work1            'and high period
        mov y, HiPer                'calculate total period
        add y, LowPer               'total period in tics
{        cmp y, C1500 wc             'see if it is reasonable
        rdword x, CAdtheta           'save in case we need to ignore this one
  if_nc jmp #UOldTh                 'zero crossing, use old theta }
        shr y, #1                   'keep within 16 bits
        mov x, HiPer                'set up dividend
        shl x, #9                   'scale by 1024 (note divisor shifted by one)
        call #divide                'calculate dc
        andn x, Hi16                'discard remainder
        sub x, DCMin                'translate to zero
        mov y, #359
        call #multiply
        mov x, y
        mov y, dcRange
        call #divide                'scale to degrees
        andn x, hi16                'drop remainder
        wrword x, CAdtheta           'save actual angle for GetPos
        mov myTheta, x              'get my copy of actual
        rdbyte work1, CAdMode wz     'check mode
   if_z rdword work1, CAdPWidth       'get current commanded pulse width (if any)
   if_z jmp #VelMd                   'we are in velocity mode       
        rdword MyCommand, CAdCommand 'get the commanded angle
        min MyCommand, #3            'set limits
        max MyCommand, #357 
        mov error, MyTheta
        sub error, MyCommand wz       'signed error
   if_Z jmp #PWGood                   'they are the same:  dead nuts on
        abs AbsError, error           'absolute value of error
        cmp AbsError, #3 wz, wc        'desired accuracy
   if_C jmp #PWGood                   'close enough:  stop servo
        or outa, Corr                   'we will correct (scope)
        mov CountGood, #0             'certainly not done
        wrword zero, CAdGoodCnt        'save in hub as well
        mov x, AbsError
        mov y, #4                      'gain
        call #multiply                 'leaves gain times error in y
        min y, #28                    'get out of dead zone
        max y, #80                   'maximum difference from 1500
        mov work1, PW1500             'get nominal pulse width         
        sub PrevTheta, MyTheta        'check for having crosswed zero
        abs PrevTheta, PrevTheta      'absolute value
        cmp PrevTheta, #300 wc, wz    'was it a lot?
  if_NC xor outa, P5                  'indicate wrong side of zero 
  if_NC xor WSZFlag, #1 wz            'invert the flag
        cmp WSZFlag, #0 wz              
  if_NZ jmp #WSZero                    'wrong side of Zero          
        cmp MyTheta, MyCommand wc     'see which way 
   if_C jmp #CW                      'current theta is less than commanded         
CCW     add work1, y                  'drive clockwise
        or outa, P3                   'indicate CW  
        jmp #HNPW
CW      sub work1, y                  'drive ccw
        jmp #HNPW
WSZero  'mov y, #30                'go slowly
        cmp MyTheta, MyCommand wc     'reverse directions
  if_NC jmp #CCW
        jmp #CW                     
PWGood  add CountGood, #1              'another good pass
        wrword CountGood, CAdGoodCnt   'save for next pass
        cmp CountGood, #10 wz, wc      'arbitrary number (200 msec)
  if_nc wrbyte zero, CAdComplete         'signal host we are done
  if_nc or outa, P4
        mov work1, PW1500              'stop servo
HNPW    wrword work1, CAdPWidth        'Have New Pulse Width: store for next pass
VelMd   mov y, work1                   'set up to convert              
        mov x, microSec              'tics in usec
        call #multiply               'convert to clock tics
        movs ctra, ContPin           'set up control pin
        mov work1, #1                '(re)enable it
        shl work1, ContPin
        or dira, work1
        andn outa, RFBMask           'end of computation                 
        waitcnt timer, C2500         '2.5 msec period
        neg phsa, y                   'start a pulse
        andn outa, Corr                 'will issue change 
        andn outa, P3                 'clockwise change
        andn outa, P4                  'high says move complete
        andn outa, P5                 'wrong side of zero
        mov PrevTheta, MyTheta        'save previous angle
        add TheServo, #1              'to next servo                        
        jmp #loop                     'forever
NoServo waitcnt timer, C2500         'just let its time be wasted
        andn outa, RFBMask           'extra long pulse
        add TheServo, #1
        jmp #loop                 

' Divide x[31..0] by y[15..0] (y[16] must be 0)
' on exit, quotient is in x[15..0] and remainder is in x[31..16]
divide  shl  y,#15    'get divisor into y[30..15]
        mov  t,#16    'ready for 16 quotient bits
:loop   cmpsub x,y wc 'y =< x? Subtract it, quotient bit in c
        rcl    x,#1    'rotate c into quotient, shift dividend
        djnz   t,#:loop 'loop until done
divide_ret ret          'quotient in x[15..0], 'remainder in x[31..16]


' Multiply x[15..0] by y[15..0] (y[31..16] must be 0)
' on exit, product in y[31..0]
multiply shl   x,#16   'get multiplicand into x[31..16]
         mov   t,#16   'ready for 16 multiplier bits
         shr   y,#1 wc 'get initial multiplier bit into c
:loop if_c add y,x wc 'if c set, add multiplicand to product
         rcr   y,#1 wc 'put next multiplier in c, shift prod.
         djnz  t,#:loop 'loop until done
multiply_ret ret 'return with product in y[31..0]         
                    
zero        long  0
C1500       long  1500          '1.5 msec 
C2500       long  2500          'number usec in 2.5 msec
RFBMask     long  %0010         'I/O 1
Hi16        long  $FFFF_0000    'high word
DcMin       long  33
DcRange     long  961
PW1500      long  1500 
Corr        long  1 << 2         'will adust pulse width
P3          long  1 << 3
P4          long  1 << 4
P5          long  1 << 5
TheServo    res                 'which servo we are doing
MicroSec    res                 'clock tics in one microsec   
AdPar       res                 'parameter list                                                     
AdPWidth    res                 'address of control pulse width in hub list
AdMode      res                 'address of mode (veleocity/angle) list
Adtheta     res                 'address of angle feedback list
AdCommand   res                 'address of commanded angle list
AdComplete  res                 'address of complete flag list
AdPins      res                 'address of pin list
AdGoodCnt   res                 'add of goodcount list
CAdPWidth   res                 'address of current control pulse width in hub
CAdMode     res                 'address of current mode (veleocity/angle)
CAdtheta    res                 'address of current angle feedback
CAdCommand  res                 'address of current commanded angle
CAdComplete  res                 'address of current complete flag
CAdPins     res                 'address of currect pins
CAdGoodCnt  res                 'address of current goodcount     
MyTheta     res                 'my copy of actual angle
MyCommand   res                 'my copy of commanded angle
Error       res                 'signed error
AbsError    res                 'absolute value of error   
HiPer       res                 'high period in clks
LowPer      res                 'low period
ContPin     res                 'control pin number
FBPinMask   res                 'feedback pin mask
timer       res                 'clock at end of next period
work1       res                 'generally useful
CPinMask    res                 'control pin mask
CountGood   res                 'count "no changes needful"
PrevTheta   res                 'theta from previous pass
OldPW       res                 'pulse width from previous pass
WSZFlag     res                 'wrong side of zero 
x           res                 'pasm divide/ multiply variables
y           res
t           res
            fit 250